-- (Tutorial by Ulf Norell - edited by Patrik Jansson.)
module CurryHoward where
⋆ = Set
{-
  The Curry-Howard isomorphism provides a "Logic" reading of types:
  p : P             p is a proof of P

  empty type        False
  non-empty type    True
  P -> Q            P implies Q
  (a : A) -> P      ∀a:A. P
  ...

-}
-- Propositions as types in Agda

data False : ⋆ where   -- empty type

data True : ⋆ where    -- non-empty type, Haskell type: ()
  unit : True          -- any constructor name would do

data _∧_ (P Q : ⋆) : ⋆ where  -- the wedge is pronounced "and"
  _,_ : P  ->  Q  ->  P ∧ Q

{-
postulate Bool : ⋆
          true : Bool
          false : Bool
testp : Bool ∧ Bool
testp = (true , false)
-}

-- A proof of (P ∧ Q) is a pair of
--   a proof of P   and
--   a proof of Q

data _∨_ (P Q : ⋆) : ⋆ where -- pronounced "or", Either p q in Haskell
  inl : P -> P ∨ Q
  inr : Q -> P ∨ Q

-- A proof of (P ∨ Q) is a either
--   a proof of P   or
--   a proof of Q

data ∃ (A : ⋆)(P : A -> ⋆) : ⋆  where
  ex : (wit : A) -> P wit -> ∃ A P

-- A proof of (∃ A P) is
--    a witness   and a proof of
--    P witness

-- Built-in:

-- ∀-quantification:
--  A proof of ((a : A) -> P a) is a function taking
--   a value a (of type A)   to
--   a proof of (P a)

-- Implication:
-- A proof of (P -> Q) is a function taking
--   a proof of P   to
--   a proof of Q

----------------
-- Negation

¬_ : ⋆ -> ⋆
¬ A  =  A -> False

test1 : ¬ False
test1 = \p -> p

test2 : ¬ ¬ True    -- (True → False) → False
test2 = \f -> f unit


-- Some simple examples

const : {A B : ⋆} -> A -> (B -> A)
const = λ x → λ y → x

swap : {P Q : ⋆} -> P ∧ Q -> Q ∧ P
swap (p , q) = (q , p)

{-
-- Bug / feature (limit / strength): the logic is _constructive_
excludedMiddle : (P : ⋆) -> (P  ∨  ¬ P)
excludedMiddle P = ? -- can not be implemented
-}

-- Next: Families.agda






-- Detail:
notNotEM : (P : ⋆) -> ¬ ¬ (P  ∨  ¬ P)
notNotEM P = \f -> f (inr (\p -> f (inl p)))
